from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.cron import CronTrigger
from fastapi import FastAPI, HTTPException, Request
from fastapi.staticfiles import StaticFiles
from loguru import logger
from starlette.responses import PlainTextResponse
from uvicorn import Config, Server

from .config import config
from .log import LOGGING_CONFIG
from .subscribe import counter, update_provider, generate_profile

if config.sentry_dsn:
    import sentry_sdk

    sentry_sdk.init(
        dsn=config.sentry_dsn,
        traces_sample_rate=1.0,
    )


app = FastAPI()

# provider download
app.mount(f"/{config.urlprefix}/provider", StaticFiles(directory="data/provider"))


# profile download
@app.get(f"/{config.urlprefix}/profile" + "/{path}")
async def profile(request: Request, path: str, id: str = None):
    path = path.rsplit(".", 1)[0] + ".yaml"

    # check profile is exists
    if not (profile_ := config.profiles.get(path[:-5])):
        raise HTTPException(404, f"Profile {path} not found")

    # check id if ids exists
    if profile_.ids and id not in profile_.ids:
        raise HTTPException(status_code=403, detail="Invalid ID")

    # logging id and ip
    user_ip = request.headers.get("X-Real-IP", request.client.host)
    logger.info(
        f"A request from {id}({user_ip}) to download profile {path} was received"
    )

    resp = PlainTextResponse(
        content=await generate_profile(path[:-5]), headers=config.headers.copy()
    )
    counter_info = await counter(path[:-5])
    if counter_info:
        resp.headers["subscription-userinfo"] = counter_info
    return resp


# manual update trigger
@app.get(f"/{config.urlprefix}/update")
async def _():
    logger.info("Update is triggered manually")
    error = await update_provider()
    return str(error) or "update complete"


# test sentry debug
@app.get("/sentry-debug")
async def trigger_error():
    division_by_zero = 1 / 0


@app.on_event("startup")
async def startup_event():
    error = await update_provider()
    if error:
        raise error
    logger.info(
        f"Starting up scheduler from crontab {config.update_cron} at timezone {config.update_tz}"
    )
    scheduler = AsyncIOScheduler()
    scheduler.add_job(
        update_provider, CronTrigger.from_crontab(config.update_cron, config.update_tz)
    )
    scheduler.start()
    logger.info(
        f"Application startup complete, listening requests from {config.domian}/{config.urlprefix}/"
    )


def main():
    Server(
        Config(
            app,
            host=config.host,
            port=config.port,
            log_config=LOGGING_CONFIG,
            reload=True,  # Enable "hot-reloading"
            reload_includes=[config.config_file_path],  # Watch 'config.yaml'
        )
    ).run()



import asyncio
from asyncio import Semaphore
from datetime import datetime
from typing import Optional

from anyio import Path
from httpx import AsyncClient
from loguru import logger

from .config import config


class Download:
    client: Optional[AsyncClient] = None
    sem: Optional[Semaphore] = None

    @classmethod
    def init(cls):
        cls.client = AsyncClient(proxies=config.download_proxy, follow_redirects=True)
        cls.sem = Semaphore(config.download_thread)

    @classmethod
    async def content(cls, url: str) -> bytes:
        if not cls.client or not cls.sem:
            cls.init()

        async with cls.sem:
            for count in range(config.download_retry):
                try:
                    res = await cls.client.get(url, timeout=60)
                    return res.content
                except Exception as e:
                    logger.error(f"[{count+1}] get {url} failed: {e}")
            logger.error(
                f"[{count+1}] {url} has reached the maximum retries, stop retries"
            )
            return b""

    @classmethod
    async def provider(cls, rulesets) -> None:
        async def download_and_write(name: str, url: str):
            path = Path(f"data/provider/{name}.yaml")
            logger.debug(f"downloading ruleset: {name}")
            bfile = await cls.content(url)
            if bfile:
                await path.write_bytes(updatetime + bfile)

        if not cls.client or not cls.sem:
            cls.init()

        tasks = []
        logger.info("Start downloading rulesets")
        updatetime = f"# update at {datetime.now()}\n".encode()
        for name in rulesets:
            url = rulesets[name]
            tasks.append(asyncio.create_task(download_and_write(name, url)))
        await asyncio.gather(*tasks)



import json
from pathlib import Path
from typing import Dict, List, Literal, Optional, Union, Set

import yaml
from pydantic import BaseModel, Extra, validator
from pytz import UnknownTimeZoneError, timezone


class Subscribe(BaseModel, extra=Extra.allow):
    type: str
    subtz: str = "Asia/Shanghai"

    @validator("subtz")
    def check_timezone(cls, tz: str):
        try:
            timezone(tz)
        except UnknownTimeZoneError as e:
            raise ValueError(f"Timezone {tz} could not be resolved") from e
        return tz


class JMS(Subscribe):
    """just my socks"""

    type: Literal["jms"] = "jms"
    url: str
    counter: Optional[str]


class ClashSub(Subscribe):
    """Generic clash profile subscription"""

    type: Literal["ClashSub"] = "ClashSub"
    url: str


class ClashFile(Subscribe):
    """Generic clash profile on local disk"""

    type: Literal["ClashFile"] = "ClashFile"
    file: str


class Profile(BaseModel):
    template: str
    subs: List[str] = []
    ids: Set[str] = set()


class Config(BaseModel, extra=Extra.ignore):
    log_level: Literal[
        "TRACE", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"
    ] = "INFO"
    sentry_dsn: Optional[str]

    download_thread: int = 4
    download_retry: int = 3
    download_proxy: Optional[str] = None

    update_cron: str = "35 6 * * *"
    update_tz: str = "Asia/Shanghai"

    domian: str = "http://0.0.0.0:46199"
    host: str = "127.0.0.1"
    port: int = 46199
    urlprefix: str = "/path/to/mess/url"
    headers: Dict[str, str] = {"profile-update-interval": "24"}

    subscribes: Dict[str, Union[JMS, ClashSub, ClashFile]]
    profiles: Dict[str, Profile]

    config_file_path: str

    @validator("port")
    def check_port(cls, p):
        if p > 65535 or p <= 0:
            raise ValueError(f"Port number must be in the range 0 to 65535, not {p}")
        return p

    @validator("update_tz")
    def check_timezone(cls, tz: str):
        try:
            timezone(tz)
        except UnknownTimeZoneError as e:
            raise ValueError(f"Timezone {tz} could not be resolved") from e
        return tz

    @validator("urlprefix", "domian", pre=True)
    def format_urlprefix(cls, v: str):
        return v.strip("/")

    @validator("profiles")
    def validate_profiles(cls, v: Dict[str, Profile], values):
        for profile in v.values():
            # check template
            if not Path(f"data/template/{profile.template}.yaml").exists():
                raise ValueError(f"template {profile.template}.yaml not exists")
            # check subscribes
            for sub in profile.subs:
                if sub not in values["subscribes"].keys():
                    raise ValueError(f"subscribe {sub} not exists")
        return v

    @staticmethod
    def _create_file(file: Path):
        file.write_text(
            Path(__file__)
            .parent.joinpath("static/config.exp.yaml")
            .read_text(encoding="utf-8"),
            encoding="utf-8",
        )

    @staticmethod
    def valueerror_parser(e: ValueError):
        return {
            ".".join([str(x) for x in err["loc"]]): err["msg"]
            for err in json.loads(e.json())
        }

    @classmethod
    def load(cls, file: Path):
        global config
        if not file.exists():
            cls._create_file(file)
            raise FileNotFoundError(
                f"Configuration file not found in {file}, a blank configuration file was created at that location"
            )
        config = cls(
            **yaml.load(file.read_bytes(), Loader=yaml.FullLoader),
            config_file_path=file.absolute().as_posix(),
        )


config: Optional[Config] = None



import base64
import json
from datetime import datetime
from typing import List, Optional, Union

from pytz import timezone

from ..model.clash import SS, Vmess
from ..utils import Download
from .base64 import ss, vmess


async def counter(url, tz: Optional[str] = None):
    tz = timezone(tz) if tz else None
    info = json.loads(await Download.content(url))
    download_ = info["bw_counter_b"]
    total = info["monthly_bw_limit_b"]
    expire_time = datetime.now(tz)

    if expire_time.day >= info["bw_reset_day_of_month"]:
        # If today's date passes reset_day_of_month
        # the next month should be taken
        if expire_time.month == 12:
            year = expire_time.year + 1
            month = expire_time.month = 1
        else:
            year = expire_time.year
            month = expire_time.month + 1
    else:
        year = expire_time.year
        month = expire_time.month

    expire = int(
        datetime(
            year=year,
            month=month,
            day=info["bw_reset_day_of_month"],
            tzinfo=tz,
        ).timestamp()
    )

    return f"upload=0; download={download_}; total={total}; expire={expire}"


async def get(url: str) -> List[Union[SS, Vmess]]:
    bsubs = await Download.content(url)
    subs = base64.decodebytes(bsubs).decode().split("\n")
    proxies = []
    for sub in subs:
        if sub.startswith("ss://"):
            proxies.append(ss(sub))
        elif sub.startswith("vmess://"):
            proxies.append(vmess(sub))
    return proxies



import base64
import json

from .clash import SS, Vmess


def ss(sub: str) -> "SS":
    """Format the ShadowSocks proxy like ss://{base64encode}#{name}@{server}:{port}"""
    ci_pa, se_po = sub[5:].split("#")

    ci_pa = base64.b64decode(f"{ci_pa}===").decode().split("@")[0]
    ci, pa = ci_pa.split(":")

    se, po = se_po.split("@")[1].split(":")
    na = "JMS-" + se.split(".")[0]

    return SS(
        **{
            "name": na,
            "server": se,
            "type": "ss",
            "port": int(po),
            "cipher": ci,
            "password": pa,
            "udp": True,
        }
    )


def vmess(sub: str) -> "Vmess":
    """Format the Vmess proxy like vmess://{base64encode}"""
    vmess = json.loads(base64.b64decode(f"{sub[8:]}===").decode())

    se = str(vmess["ps"]).split("@")[1].split(":")[0]
    na = "JMS-" + se.split(".")[0]

    return Vmess(
        **{
            "name": na,
            "server": se,
            "port": int(vmess["port"]),
            "type": "vmess",
            "uuid": vmess["id"],
            "alterId": vmess["aid"],
            "cipher": "auto",
            "tls": vmess["tls"] != "none",
            "skip-cert-verify": True,
            "udp": True,
        }
    )



from typing import Literal, Optional

from pydantic import BaseModel, Extra, Field, validator


class SS(BaseModel, extra=Extra.allow):
    name: str
    type: Literal["ss"] = "ss"
    server: str
    port: int
    cipher: Literal[
        "aes-128-gcm",
        "aes-192-gcm",
        "aes-256-gcm",
        "aes-128-cfb",
        "aes-192-cfb",
        "aes-256-cfb",
        "aes-128-ctr",
        "aes-192-ctr",
        "aes-256-ctr",
        "rc4-md5",
        "chacha20-ietf",
        "xchacha20",
        "chacha20-ietf-poly1305",
        "xchacha20-ietf-poly1305",
    ]
    password: str
    udp: Optional[bool] = False


class SSR(SS):
    type: Literal["ssr"] = "ssr"
    obfs: Literal[
        "plain",
        "http_simple",
        "http_post",
        "random_head",
        "tls1.2_ticket_auth",
        "tls1.2_ticket_fastauth",
    ]
    protocol: Literal[
        "origin",
        "auth_sha1_v4",
        "auth_aes128_md5",
        "auth_aes128_sha1",
        "auth_chain_a",
        "auth_chain_b",
    ]
    obfs_param: Optional[str] = Field(default=None, alias="obfs-param")
    protocol_param: Optional[str] = Field(default=None, alias="protocol-param")
    udp: Optional[bool] = False


class Vmess(BaseModel, extra=Extra.allow):
    name: str
    type: Literal["vmess"] = "vmess"
    server: str
    port: int
    uuid: str
    alterId: int
    cipher: Literal["auto", "aes-128-gcm", "chacha20-poly1305", "none"] = "auto"


class Socks5(BaseModel, extra=Extra.allow):
    name: str
    type: Literal["socks5"] = "socks5"
    server: str
    port: int


class Snell(BaseModel, extra=Extra.allow):
    name: str
    type: Literal["snell"] = "snell"
    server: str
    port: int
    psk: str


class Trojan(BaseModel, extra=Extra.allow):
    name: str
    type: Literal["trojan"] = "trojan"
    server: str
    port: int
    password: str



from pathlib import Path
from typing import Dict, List, Literal, Optional, Union, Sequence

import yaml
from pydantic import BaseModel, Extra, Field, validator

from .proxy import SS, SSR, Snell, Socks5, Trojan, Vmess
from .proxygroup import ProxyGroup, ProxyGroupTemplate
from .ruleprovider import RuleProvider


class ClashTemplate(BaseModel, extra=Extra.allow):
    proxies: Union[
        List[Union[SS, SSR, Vmess, Socks5, Snell, Trojan]], Literal["__proxies_list__"]
    ]
    proxy_groups: List[ProxyGroupTemplate] = Field(alias="proxy-groups")
    rule_providers: Optional[Dict[str, RuleProvider]] = Field(
        default=None, alias="rule-providers"
    )
    rules: List[str]

    @classmethod
    def load(cls, file: str) -> "ClashTemplate":
        return cls(
            **yaml.load(
                Path(f"data/template/{file}.yaml").read_bytes(), Loader=yaml.FullLoader
            )
        )

    def render(
        self, proxies: List[Union[SS, SSR, Vmess, Socks5, Snell, Trojan]]
    ) -> "Clash":
        if not proxies:
            return Clash(**self.dict(exclude_none=True, by_alias=True))
        proxies_name_list = [proxy.name for proxy in proxies]
        proxy_groups = []
        for group in self.proxy_groups:
            if group.proxies == "__proxies_name_list__":
                group.proxies = proxies_name_list
            if (
                isinstance(group.proxies, Sequence)
                and "__proxies_name_list__" in group.proxies
            ):
                group.proxies.remove("__proxies_name_list__")
                group.proxies += proxies_name_list
            proxy_groups.append(group)

        self.proxies = proxies
        self.proxy_groups = proxy_groups
        return Clash(**self.dict(exclude_none=True, by_alias=True))


class Clash(ClashTemplate):
    proxies: List[Union[SS, SSR, Vmess, Socks5, Snell, Trojan]]
    proxy_groups: List[ProxyGroup] = Field(alias="proxy-groups")



import logging
import sys
from pathlib import Path

from loguru import logger

from .config import config

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {
        "default": {
            "class": "LiteClashProMan.log.LoguruHandler",
        },
    },
    "loggers": {
        "uvicorn.error": {"handlers": ["default"], "level": "INFO"},
        "uvicorn.access": {
            "handlers": ["default"],
            "level": "INFO",
        },
    },
}


class LoguruHandler(logging.Handler):  # pragma: no cover
    """logging 与 loguru 之间的桥梁，将 logging 的日志转发到 loguru。"""

    def emit(self, record: logging.LogRecord):
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        frame, depth = logging.currentframe(), 2
        while frame and frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(
            level, record.getMessage()
        )


logger.remove()

# ensure logs dir exist
LOGPATH = Path("logs")
LOGPATH.mkdir(exist_ok=True)
LOGURU_FORMAT = "<green>{time:YYYY-MM-DD HH:mm:ss}</green>|<level>{level: <8}</level>|\
<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>"


# add file logger
logger.add(
    LOGPATH.joinpath("{time:YYYY-MM-DD}.log"),
    format=LOGURU_FORMAT,
    encoding="utf-8",
    backtrace=True,
    diagnose=True,
    rotation="00:00",
    colorize=False,
    level="DEBUG",
)

# add stdout logger
logger.add(
    sys.stdout,
    format=LOGURU_FORMAT,
    backtrace=True,
    diagnose=True,
    colorize=True,
    level=config.log_level,
)



from pathlib import Path
from typing import List, Union

import yaml
from loguru import logger

from ..config import config
from ..model.clash import SS, SSR, ClashTemplate, Snell, Socks5, Trojan, Vmess
from ..utils import Download
from . import clash, jms


async def _subs(
    subs: List[str],
) -> List[Union[SS, SSR, Vmess, Socks5, Snell, Trojan]]:
    proxies = []
    for name in subs:
        sub = config.subscribes[name]
        if sub.type == "jms":
            proxies += await jms.get(sub.url)
        if sub.type == "ClashSub":
            proxies += await clash.get_sub(sub.url)
        if sub.type == "ClashFile":
            proxies += await clash.get_file(sub.file)
    return proxies


async def generate_profile(profile: str):
    proxies = await _subs(config.profiles[profile].subs)
    logger.debug(
        f"Generating profile {profile} from template {config.profiles[profile].template}"
    )
    template = ClashTemplate.load(config.profiles[profile].template)
    clash = template.render(proxies)
    if clash.rule_providers:
        for provider in clash.rule_providers:
            # if provider is exists in local
            # replace it with loacl file
            if Path(f"data/provider/{provider}.yaml").exists():
                clash.rule_providers[provider].url = "/".join(
                    [
                        config.domian,
                        config.urlprefix,
                        "provider",
                        f"{provider}.yaml",
                    ]
                )
    return yaml.dump(
        clash.model_dump(exclude_none=True, by_alias=True, exclude_unset=True),
        sort_keys=False,
        allow_unicode=True,
    )


async def update_provider():
    logger.info("Start update provider")
    try:
        rulesets = {}
        for profile in config.profiles:
            template = ClashTemplate.load(config.profiles[profile].template)
            if template.rule_providers:
                for provider in template.rule_providers:
                    rulesets[provider] = template.rule_providers[provider].url
        await Download.provider(rulesets)
        logger.success("Update complete")

    except Exception as e:
        logger.critical(e)
        return e


async def counter(profile: str):
    if len(config.profiles[profile].subs) != 1:
        logger.warning("Subscribe(s) is not 1, counter function disabled")
        return ""
    sub = config.subscribes[config.profiles[profile].subs[0]]
    if sub.type == "jms" and sub.counter:
        return await jms.counter(sub.counter, sub.subtz)
    elif sub.type == "ClashSub":
        return await clash.counter(sub.url)
    return ""



import argparse
import shutil
from pathlib import Path

from .config import Config

parser = argparse.ArgumentParser(description="run LiteClashProMan")
parser.add_argument(
    "--config", "-c", default="config.yaml", help="Path to the config file"
)
args = parser.parse_args()


def main():
    # check dirs and files
    static_dir = Path(__file__).parent.joinpath("static")
    data_dir = Path("data")
    provider_dir = data_dir.joinpath("provider")
    template_dir = data_dir.joinpath("template")

    for dir in [data_dir, provider_dir, template_dir]:
        if not dir.exists():
            dir.mkdir(0o755, parents=True, exist_ok=True)
        assert dir.is_dir(), f"{dir.as_posix()} should be folder"

    if not any(template_dir.glob("*")):
        shutil.copytree(
            static_dir.joinpath("template"), template_dir, dirs_exist_ok=True
        )

    Config.load(Path(args.config))

    from .main import main as run_main

    run_main()



from pathlib import Path
from typing import List, Union

import yaml
from httpx import Response

from ..model.clash import SS, SSR, Clash, Snell, Socks5, Trojan, Vmess
from ..utils import Download


async def counter(url):
    resp: Response = await Download.client.get(url)
    try:
        return resp.headers.get("subscription-userinfo")
    except KeyError:
        return ""


async def get_sub(url: str) -> List[Union[SS, SSR, Snell, Socks5, Trojan, Vmess]]:
    return Clash(
        **yaml.load(await Download.content(url), Loader=yaml.FullLoader)
    ).proxies


async def get_file(file: str) -> List[Union[SS, SSR, Snell, Socks5, Trojan, Vmess]]:
    return Clash(**yaml.load(Path(file).read_bytes(), Loader=yaml.FullLoader)).proxies



from typing import Literal, Optional, Union, List

from pydantic import BaseModel, Extra


class ProxyGroupTemplate(BaseModel, extra=Extra.allow):
    name: str
    type: str
    proxies: Union[List[str], Literal["__proxies_name_list__"]]
    url: Optional[str] = None
    interval: Optional[int] = None
    _index: Optional[int] = None


class ProxyGroup(ProxyGroupTemplate):
    proxies: List[str]
